/*
 * Butterworth and Chebyshev filter poles routines.
 *
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "macros.h"
#include "common.h"
#include "complex.h"
#include "util.h"
#include "landen_trans.h"

/*
 * filter_make_spec - creates filter specification using parameters.
 *
 * type:		type of the filter (lowpass, highpass)
 * approx:		type of approximation polynomial
 * ripple:		ripple in passband in dB
 * attenuation:		attenuation in the stop band in dB
 * freq_cutoff:		cutoff frequency, i.e. 3dB attenuation point
 * freq_stop:           stop band frequency
 *
 * Returns: filter specification structure or NULL if invalid parameters.
 *
 * All frequencies are in radians. Use 2.0*M_PI factor if frequencies are in Hz.
 */

struct filter_spec *
filter_make_spec(enum filter_type type, enum filter_approx approx, double ripple, double atten_stop, double freq_pass1, double freq_stop1, double freq_pass2, double freq_stop2)
{
	struct filter_spec *spec;
	double fp, fs, fp1, fs1, fp2, fs2;
	double omeg, w, eps_p, eps_s, k, k1;
	double num, den;
	int i;
	unsigned int n;

	fp = fs = fs1 = fs2 = 0.0;
	omeg = w = eps_p = eps_s = k = k1 = 0.0;
	num = den = 0.0;

	if (atten_stop <= 0.0)
		return NULL;

	switch (type) {
	case LOWPASS:
		fp = 1.0;
		fs = freq_stop1 / freq_pass1;
		break;
	case HIGHPASS:
		fp = 1.0;
		fs = freq_pass1 / freq_stop1;
		break;
	case BANDPASS:
		fp = 1.0;
		omeg = freq_pass1 * freq_pass2;
		w = freq_pass2 - freq_pass1; 
		fs1 = fabs((freq_stop1 * freq_stop1 - omeg) / (w * freq_stop1)); 
		fs2 = fabs((freq_stop2 * freq_stop2 - omeg) / (w * freq_stop2));
		if (fs1 < fs2)
			fs = fs1; 
		else 
			fs = fs2;
		break; 
	case BANDSTOP:
		fp = 1.0;
		omeg = freq_pass1 * freq_pass2;
		w = freq_pass2 - freq_pass1;
		fs1 = fabs((w * freq_stop1) / (omeg - freq_stop1 * freq_stop1)); 
		fs2 = fabs((w * freq_stop2) / (omeg - freq_stop2 * freq_stop2)); 		
		if (fs1 < fs2)
			fs = fs1; 
		else
			fs = fs2;
		break; 		    
	}

	if ((fp <= 0.0) || (fs <= fp))
		return NULL;

	switch (approx) {
	case BUTTERWORTH:
		num = log10(pow(10.0, 0.1 * atten_stop) - 1.0);
		den = 2.0 * log10(fs / fp);
		ripple = 3.0;
		break;
	case CHEBYSHEV:
		eps_p = sqrt(pow(10.0, ripple / 10.0) - 1.0);
		eps_s = sqrt(pow(10.0, atten_stop / 10.0) - 1.0);
		num = acosh(eps_s / eps_p);
		den = acosh(fs / fp);
		break;
	case ELLIPTIC:
		eps_p = sqrt(pow(10.0, ripple / 10.0) - 1.0);
		eps_s = sqrt(pow(10.0, atten_stop / 10.0) - 1.0);
		k = fp / fs;
		k1 = eps_p / eps_s;
		num = complete_intgr(k) * complement_intgr(k1);
		den = complete_intgr(k1) * complement_intgr(k);
		
		k = recalc_k(ceil(num / den), k1);
		break;
	}

	n = ceil(num / den);
	
	OBJ_NEW(spec);

	spec->type        = type;
	spec->order       = n;
	spec->atten_pass  = ripple;
	spec->atten_stop  = atten_stop;
	spec->freq_pass1  = freq_pass1;
	spec->freq_stop1  = freq_stop1;
	spec->freq_pass2  = freq_pass2;
	spec->freq_stop2  = freq_stop2;	
	spec->approx      = approx;
	spec->k		  = k ;
	
	return spec;
}

/*
 * filter_analog - creates analog filter transfer function H(s) using specification.
 *
 * spec:		pointer to filter specification
 *
 * Returns: transfer function H(s) in the s-plane.
 */

struct transfer_func *
filter_analog(struct filter_spec *spec)
{
	struct transfer_func *h;
	double a, b, c;
	double epsilon, gamma;
	double epsilon_p, epsilon_s, k1, w, omeg, sqrt_omeg;
	int i, rounds, n;
	complex_t *poles;
	complex_t *zeros;
	complex_t p, s;
	complex_t u, cd_k, j, v0;
	complex_t sqr_complex, zero_square, pole_square;
	complex_t complex_a, complex_b;
	
	return_val_if_fail(spec != NULL, NULL);

	n = spec->order;
	
	OBJ_NEW(h);
	
	h->h_zero = 1.0;
	
	zeros = xmalloc0(n * sizeof(*h->zeros));
	poles = xmalloc0(n * sizeof(*h->poles));
	
	switch (spec->approx) {
	case BUTTERWORTH:
		for (i = 0; i < n; i++)
			zeros[i] = c_num(0.0, INFINITY);
		
		for (i = 0; i < (n / 2); i++) {
			a = M_PI * ((2.0 * (i + 1) + (n - 1)) / (2.0 * n));
			p = c_num(cos(a), sin(a));
			poles[2 * i] = p;
			poles[2 * i + 1] = c_conjugate(p);
		}

		if (n % 2)
			poles[n - 1] = c_num(-1.0, 0.0);
			
		break;
	case CHEBYSHEV:
		epsilon = sqrt(pow(10.0, spec->atten_pass / 10.0) - 1.0);
		gamma = (1.0 + sqrt(1.0 + epsilon * epsilon)) / epsilon;
		gamma = pow(gamma, 1.0 / n);
		a = ((1.0 / gamma) - gamma) / 2.0;
		b = ((1.0 / gamma) + gamma) / 2.0;

		
		for (i = 0; i < n; i++)
			zeros[i] = c_num(0.0, INFINITY);
			
		if (n % 2) {
			rounds = (n / 2) + 1;

			for (i = 0; i < rounds; i++) {
				c = ((2 * (i + 1) - 1) * M_PI) / (2.0 * n);
				s.re = a * sin(c);
				s.im = b * cos(c);
				if (i < rounds - 1) {
					poles[2 * i] = s;
					poles[2 * i + 1] = c_conjugate(s);
				} else {
					poles[n - 1].re = s.re;
					poles[n - 1].im = 0.0;
				}
			}

			/* hzero is calculated using pole conjugate pairs */
			p = c_num(-1.0, 0.0);

			for (i = 0; i < n / 2; i++)
				p = c_mult(p, c_mult(poles[2 * i], poles[2 * i + 1]));

			h->h_zero = p.re * s.re;
			
		} else {
			rounds = n / 2;

			for (i = 0; i < rounds; i++) {
				c = ((2.0 * (i + 1) - 1.0) / (2.0 * n)) * M_PI;
				s.re = a * sin(c);
				s.im = b * cos(c);
				poles[2 * i] = s;
				poles[2 * i + 1] = c_conjugate(s);
			}

			/* All poles form conjugate pairs. Nothing to worry about. */
			p = c_num(1.0, 0.0);

			for (i = 0; i < n / 2; i++)
				p = c_mult(p, c_mult(poles[2 * i], poles[2 * i + 1]));

			/* DC gain is 1/(1+epsilon^2)^1/2 */
			
			h->h_zero = p.re * (1.0 / sqrt(1.0 + epsilon * epsilon));
		}

		/* Normalize cutoff to 3dB. */

		/* calculate A */
		a = acosh(1.0 / epsilon) / (double)n;
		/* calculate R */
		b = cosh(a);
		c = 1.0;
		
		for (i = 0; i < n; i++) {
			poles[i] = c_scale(poles[i], 1.0 / b);
			c *= b; /* c = R^npoles */
		}
		
		h->h_zero /= c;
		break;
	case ELLIPTIC:
		epsilon_p = sqrt(pow(10.0, spec->atten_pass / 10.0) - 1.0);
		epsilon_s = sqrt(pow(10.0, spec->atten_stop / 10.0) - 1.0);
		k1 = epsilon_p / epsilon_s;
		j = c_num(0.0, 1.0);
		complex_a = sn_inverse(c_num(0.0, 1.0 / epsilon_p), k1);
		complex_b = c_num((double)n, 0.0);
		v0 = c_div(complex_a, complex_b);
		
		/* Calculate zeros */
		for (i = 0; i < (n / 2); i++) {
			u = c_num((2.0 * (i + 1) - 1.0) / n, 0.0);
			cd_k = c_mult(c_num(spec->k, 0.0), cd(u, spec->k));
		
			zeros[2 * i] = c_div(j, cd_k);
			zeros[2 * i + 1] = c_conjugate(zeros[2 * i]);
		}

		if (n % 2)
			zeros[n - 1] = c_num(0.0, INFINITY);
		
		/* Calculate poles */
		for (i = 0; i < (n / 2); i++) {
			u = c_num((2.0 * (i + 1) - 1.0) / n, 0.0);
			cd_k = cd(c_sub(u, v0), spec->k);
			
			poles[2 * i] = c_mult(j, cd_k);
			poles[2 * i + 1] = c_conjugate(poles[2 * i]);
		}

		if (n % 2)
			poles[n - 1] = c_mult(j, sn(v0, spec->k));
		
		break;
	}
	
	switch (spec->type) {
	case LOWPASS:
		h->nz = h->np = spec->order;
		h->zeros = xmalloc0(h->nz * sizeof(*h->zeros));
		h->poles = xmalloc0(h->np * sizeof(*h->poles));
		
		for (i = 0; i < h->nz; i++) {
			if (zeros[i].im == INFINITY)
				h->zeros[i] = zeros[i];
			else 
				h->zeros[i] = c_scale(zeros[i], spec->freq_pass1);
		}
		
		for (i = 0; i < h->np; i++)
			/* Scale each pole to LP filter. */
			h->poles[i] = c_scale(poles[i], spec->freq_pass1);
		
		/* H0 = cutoff^order */
		if (spec->order > 1) {
			for (i = 0; i < spec->order; i++)
				h->h_zero *= spec->freq_pass1;
		} else
			h->h_zero = spec->freq_pass1;
		break;
	case HIGHPASS:
		/* Move zeros from infinity to 0. */
		h->nz = h->np = spec->order;
		h->zeros = xmalloc0(h->nz * sizeof(*h->zeros));
		h->poles = xmalloc0(h->np * sizeof(*h->poles));
		
		for (i = 0; i < h->nz; i++) {
			if (zeros[i].im == INFINITY)
				h->zeros[i] = c_num(0.0, 0.0);
			else {
				h->zeros[i] = c_inverse(zeros[i]);
				h->zeros[i] = c_mult(h->zeros[i], c_num(spec->freq_pass1, 0.0));
			}
		}
		
		for (i = 0; i < h->np; i++) {
			/* Transform LP poles to HP poles. */
			h->poles[i] = c_inverse(poles[i]);
			/* Scale each pole to cutoff frequency. */
			h->poles[i] = c_mult(h->poles[i], c_num(spec->freq_pass1, 0.0)); 
		}

		if (spec->order > 1) {
			for (i = 0; i < spec->order; i++)
				h->h_zero *= spec->freq_pass1;
		} else
			h->h_zero = spec->freq_pass1;
		break;
	case BANDPASS:
		w = spec->freq_pass2 - spec->freq_pass1;
		omeg = spec->freq_pass1 * spec->freq_pass2;
		complex_a = c_num(4.0 * omeg / (w * w), 0.0);
		complex_b = c_num(w / 2.0, 0.0);
		
		h->nz = h->np = (spec->order * 2);
		h->zeros = xmalloc0(h->nz * sizeof(*h->zeros));
		h->poles = xmalloc0(h->np * sizeof(*h->poles));	

		for (i = 0; i < (h->nz / 2); i++) {
			if (zeros[i].im == INFINITY) {
				h->zeros[i * 2] = c_num(0.0, 0.0);
				h->zeros[i * 2 + 1] = c_num(0.0, INFINITY);
			} else {
				zero_square = c_mult(zeros[i], zeros[i]);
				sqr_complex = c_sqrt(c_sub(zero_square, complex_a));
				h->zeros[i * 2] = c_mult(complex_b, (c_sum(zeros[i], sqr_complex))); 
				h->zeros[i * 2 + 1] = c_mult(complex_b, (c_sub(zeros[i], sqr_complex)));
			}
		}
		
		for (i = 0; i < (h->np / 2); i++) {
			pole_square = c_mult(poles[i], poles[i]);
			sqr_complex = c_sqrt(c_sub(pole_square, complex_a));
			h->poles[i * 2]   = c_mult(complex_b, (c_sum(poles[i], sqr_complex))); 
			h->poles[i * 2 + 1] = c_mult(complex_b, (c_sub(poles[i], sqr_complex)));
		}
		
		if (spec->order > 1) {
			for (i = 0; i < spec->order; i++)
				h->h_zero *= w;
		} else
			h->h_zero = w;
		break;
	case BANDSTOP:
		w = spec->freq_stop2 - spec->freq_stop1;
		omeg = spec->freq_stop1 * spec->freq_stop2;
		complex_a = c_num(4.0 * omeg / (w * w), 0.0);
		complex_b = c_num(w / 2.0, 0.0);
		sqrt_omeg = sqrt(omeg);

		h->nz = h->np = (spec->order * 2);
		h->zeros = xmalloc0(h->nz * sizeof(*h->zeros));
		h->poles = xmalloc0(h->np * sizeof(*h->poles));

		for (i = 0; i < (h->nz / 2); i++) {
			if (zeros[i].im == INFINITY) {
				h->zeros[i * 2] = c_num(0.0, sqrt_omeg);
				h->zeros[i * 2 + 1] = c_num(0.0, -1.0 * sqrt_omeg);
			} else {
				zeros[i] = c_inverse(zeros[i]);
				zero_square = c_mult(zeros[i], zeros[i]);		
				sqr_complex = c_sqrt(c_sub(zero_square, complex_a));
				h->zeros[i * 2]   = c_mult(complex_b, (c_sum(zeros[i], sqr_complex))); 
				h->zeros[i * 2 + 1] = c_mult(complex_b, (c_sub(zeros[i], sqr_complex)));
			}
		}

		for (i = 0; i < (h->np / 2); i++) {
			poles[i] = c_inverse(poles[i]);
			pole_square = c_mult(poles[i], poles[i]);
			sqr_complex = c_sqrt(c_sub(pole_square, complex_a));
			h->poles[i * 2]   = c_mult(complex_b, (c_sum(poles[i], sqr_complex))); 
			h->poles[i * 2 + 1] = c_mult(complex_b, (c_sub(poles[i], sqr_complex)));
		}
		
		if (spec->order > 1) {
			for (i = 0; i < spec->order; i++)
				h->h_zero *= w;
		} else
			h->h_zero = w;
		break;
	}
	
	return h;
}

